import importlib
import logging
import logging.config
from collections import OrderedDict
import yaml
from . import context
from .settings import SETTINGS
from .web.rpc_websocket import RpcHandler, RpcWebsocket
from .tasks.utils import Colored
from .tasks.utils import pid_file
from .utils.describe import Procedure

log = logging.getLogger(__name__)


class Bluemax:
    """
    Bring up in order:

        extend_settings
        extend_logging
        extend_urls

    then:

        Web
        Services
        Worker
    """

    def __init__(self, module=None, settings=None, conf=None):
        self._procedures = None
        self._services = None
        self._application = None
        self._broadcaster = None
        self._urls = None
        self.settings = None

        if conf:
            with open(conf, "r") as file:
                settings = yaml.safe_load(file)

        self.settings = self.extend_settings(settings if settings else {})

        logging_config = self.extend_logging(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "formatters": {
                    "standard": {
                        "format": "%(asctime)s %(levelname)s %(name)s: %(message)s"
                    }
                },
                "handlers": {
                    "default": {
                        "level": self.settings.get("logging", "INFO"),
                        "formatter": "standard",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",  # Default is stderr
                    }
                },
                "loggers": {
                    "": {"handlers": ["default"], "propagate": True}  # root logger
                },
            }
        )
        if logging_config:
            logging.config.dictConfig(logging_config)

        if module is None:
            module = self.settings.get("procedures")
        assert module, "module required"
        mod = self._get_module_(module)
        if mod:
            self.settings.update(mod)
        if self.settings.get("log_extend"):
            logging_config = self._qname_to_class_(self.settings["log_extend"])(
                logging_config
            )
            if logging_config:
                logging.config.dictConfig(logging_config)
        if self.settings.get("settings_extend"):
            self.settings = self._qname_to_class_(self.settings["settings_extend"])(
                self.settings
            )
        SETTINGS.update(self.settings)

    @classmethod
    def describe(cls, target):
        """
            Will inspect each function in target.__all__
            and return an instance of Procedure for each
        """
        result = OrderedDict()
        for key in target.__all__:
            f = getattr(target, key)
            s = Procedure(key, f)
            log.debug(s)
            result[key] = s
        return result

    @classmethod
    def _qname_to_class_(cls, qname):
        """ returns a class from module_name:class_name """
        if isinstance(qname, str):
            _modules, _class = qname.split(":")
            return getattr(importlib.import_module(_modules), _class)
        return qname

    @classmethod
    def _import_module_(cls, name):
        "Its ok not to have the module but not ok to not have dependants."
        try:
            logging.debug("looking for module %s", name)
            return importlib.import_module(name)
        except ModuleNotFoundError as ex:
            if name not in str(ex):
                raise

    def _get_module_(self, name):
        """ import the project and return configuration options """

        options = {}
        if name is None:
            print("module expected")
        else:
            importlib.import_module(name)
            self._procedures = importlib.import_module(f"{name}.procedures")
            if self._procedures:
                if not getattr(self._procedures, "__all__"):
                    print(f"expected __all__ in procedures module")
                else:
                    options["procedures"] = f"{name}.procedures"
                    ext_log = self._import_module_(f"{name}.log")
                    if ext_log:
                        options["log_extend"] = f"{name}.log:extend"
                    ext_settings = self._import_module_(f"{name}.settings")
                    if ext_settings:
                        options["settings_extend"] = f"{name}.settings:extend"
                    ext_urls = self._import_module_(f"{name}.urls")
                    if ext_urls:
                        options["urls_extend"] = f"{name}.urls:extend"
                    services = self._import_module_(f"{name}.services")
                    if services:
                        options["services"] = f"{name}.services"
                        self._services = services
            else:
                print(f"expected procedures in {name}")

        return options

    def make_app(self, with_services=False, loop=None):
        """ returns a tornado Application configured with urls and settings"""
        from tornado.web import Application, StaticFileHandler
        from bluemax.web.authentication import LogoutHandler
        from bluemax.web.auth_static_file_handler import AuthStaticFileHandler

        settings = {}
        settings.update(self.settings["tornado"])

        default_manager = (
            "bluemax.services:ServicesRpc"
            if with_services
            else "bluemax.rpc:ContextRpc"
        )
        manager = self._qname_to_class_(settings.get("manager", default_manager))

        settings["manager"] = manager.rpc_for(
            self.settings.get("procedures"),
            services=self.settings.get("services"),
            loop=loop,
        )
        routes = [(r"/rpc", RpcHandler), (r"/ws", RpcWebsocket)]
        if settings.get("auth_handler"):
            settings.setdefault("cookie_name", "bluemax-user")
            settings.setdefault(
                "cookie_secret", "<change me - it was a dark and stormy..>"
            )
            auth_handler = self._qname_to_class_(settings.get("auth_handler"))
            settings["login_url"] = "/login"
            routes.insert(0, (r"/logout", LogoutHandler))
            routes.insert(0, (r"/login", auth_handler))
            if settings.get("static_dir"):
                # we use this because out static_dir is root for spa
                # static_path created a static by tornado
                routes.append(
                    (
                        r"/(.*)",
                        AuthStaticFileHandler,
                        {
                            "path": settings.get("static_dir"),
                            "default_filename": "index.html",
                        },
                    )
                )
                log.info("serving from %s", settings.get("static_dir"))
        elif settings.get("static_dir"):
            # we use this because out static_dir is root for spa
            # static_path created a static by tornado
            routes.append(
                (
                    r"/(.*)",
                    StaticFileHandler,
                    {
                        "path": settings.get("static_dir"),
                        "default_filename": "index.html",
                    },
                )
            )
            log.info("serving from %s", settings.get("static_dir"))
        self._urls = self.extend_urls(routes)
        if self.settings.get("urls_extend"):
            self._urls = self._qname_to_class_(self.settings["urls_extend"])(self._urls)
        log.debug("tornado urls: %s", self._urls)
        log.debug("tornado settings: %s", settings)
        return Application(self._urls, **settings)

    def stand_alone(self):
        """
        run web, services and procedures, no redis
        """
        default_manager = (
            "bluemax.services:ServicesRpc"
            if self._services
            else "bluemax.rpc:ContextRpc"
        )
        self.settings.setdefault("tornado", {}).setdefault("manager", default_manager)
        self.web(with_services=True)

    def stand_two(self):
        """
        run web, services and procedures but can scale
        the broadcaster uses redis to connect all websockets.
        """
        default_manager = (
            "bluemax.services:ServicesRpc"
            if self._services
            else "bluemax.rpc:ContextRpc"
        )
        self.settings.setdefault("tornado", {}).setdefault("manager", default_manager)
        self.settings.setdefault(
            "broadcaster", "bluemax.work.redis_broadcaster:RedisBroadcaster"
        )
        self.web(with_services=True)

    def web(self, with_services=False, pid="server.pid"):
        """
        runs tornado application, awaiting KeyboardInterrupt
        called directly assumes redis_url
        """
        from tornado.ioloop import IOLoop, PeriodicCallback

        self.settings.setdefault("tornado", {}).setdefault(
            "manager", "bluemax.work:RedisManager"
        )
        self._application = self.make_app(with_services=with_services)

        port = self.settings.get("bluemax", {}).get("port", "8080")
        address = self.settings.get("bluemax", {}).get("address", "127.0.0.1")
        self._application.listen(int(port), address)
        log.info("listening on http://%s:%s", address, port)

        self._broadcaster = self._qname_to_class_(
            self.settings.get("broadcaster", "bluemax.web.broadcaster:Broadcaster")
        )()
        self._broadcaster.start(
            self._application.settings["manager"].get_broadcast_queue(),
            RpcWebsocket.broadcast,
        )

        ioloop = IOLoop.current()
        keep_alive_interval = int(self.settings.get("keep_alive_interval", "30000"))
        PeriodicCallback(RpcWebsocket.keep_alive, keep_alive_interval).start()

        with pid_file(pid):
            try:
                ioloop.start()
            except KeyboardInterrupt:
                log.info("stopping")
                ioloop.stop()

    def _run_loop_(self, loop, manager):
        """ run an async ioloop """
        try:
            loop.run_forever()
        finally:
            log.info("stopping")
            loop.run_until_complete(manager.shutdown())
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    def services(self, pid="services.pid"):
        """ run just services """
        import asyncio
        from bluemax.services.server import Services

        assert self.settings.get("redis_url"), "required to distrubute broadcasts"
        manager = Services(self._services)
        context._MANAGER_ = manager  # pylint: disable=W0212
        loop = asyncio.get_event_loop()
        asyncio.ensure_future(manager.start_all_services())
        with pid_file(pid):
            self._run_loop_(loop, manager)

    def worker(self, pid="worker.pid"):
        """ run just procedures """
        import asyncio
        from bluemax.work.worker import Worker

        assert self.settings.get("redis_url"), "required to distrubute work queue"
        manager = Worker.rpc_for(
            self.settings["procedures"],
            max_workers=self.settings.get("max_workers"),
            max_threads=self.settings.get("max_threads"),
        )
        loop = asyncio.get_event_loop()
        with pid_file(pid):
            self._run_loop_(loop, manager)

    def config(self):
        """ displays resultant config """
        log.info("\n---\n%s", yaml.dump(self.settings))
        print(Colored.blue("available:"))
        for description in self.describe(self._procedures).values():
            print("  ", description)
        if self._services:
            print(Colored.blue("services:"))
            for description in self.describe(self._services).values():
                print("  ", description)
        print("", flush=True)

    def extend_logging(self, config):
        """ overide this to extend logging """
        return config

    def extend_settings(self, settings):
        """ overide this to extend settings """
        return settings

    def extend_urls(self, urls):
        """ overide this to extend urls """
        return urls



"""
    Returns the current user object as json
"""
import tornado.web
from bluemax.utils import json_utils
from .user_mixin import UserMixin


class ProfileHandler(UserMixin, tornado.web.RequestHandler):  # pylint: disable=W0223
    """ this assumes the user is logged in """

    @tornado.web.authenticated
    async def get(self):
        """ return a json view of the current user """
        user = self.current_user
        self.set_header("content-type", "application/json")
        self.write(json_utils.dumps(user, indent=4))



'''
    We want multiple ECS tasks and not to have to provide
    a redis service - can aws topics and queues to the
    job
    TODO - just an idea - do not use!
'''
import asyncio
import logging
import os
from tornado.httpclient import AsyncHTTPClient
from bluemax.web.broadcaster import Broadcaster
from bluemax.utils import json_utils
from .aws_pubsub import aws_subscribe, aws_publish

LOGGER = logging.getLogger(__name__)


async def gen_queue_name():
    '''
        We need a subscription per fargate task!
        only unique this is ip address ? task arn ?
        mangle that into a name, create queue and subscriptions?
        but then how do you tidy up?
    '''
    task_uri = os.getenv('ECS_CONTAINER_METADATA_URI', None)
    if task_uri:
        http_client = AsyncHTTPClient()
        try:
            response = await http_client.fetch('http://www.google.com')
        except Exception:
            LOGGER.exception('Error: %s', task_uri)
            raise
        else:
            task_def = json_utils.loads(response.body)
            task_arn = task_def.get('TaskARN')
            LOGGER.info('task arn: %s', task_arn)


class AwsBroadcaster(Broadcaster):
    ''' Broadcast to topic and subscribe to queue '''

    def __init__(self, topic_name, queue_name=None):
        self.topic_name = topic_name
        self.queue_name = queue_name
        self.consumer = None
        super().__init__()

    async def broadcaster(self, queue, consumer):
        ''' Connect websockets to manager broadcast queue '''
        self.consumer = consumer
        LOGGER.info('setting up aws publisher')
        aws_publish(self._publish_, topic_name=self.topic_name)
        LOGGER.info('subscribing to aws queue')
        aws_subscribe(self.handle_broadcast, queue_name=self.queue_name)
        self.handle_broadcast()
        LOGGER.info('listening to local queue')
        while True:
            value = await queue.get()
            self._publish_(*value)
            queue.task_done()

    def _publish_(self, signal, message, client_filter=None):
        ''' will broadcast to aws the result '''
        msg = json_utils.dumps((signal, message, client_filter))
        LOGGER.debug('broadcast to aws: %s', msg)
        return msg

    async def handle_broadcast(self, message=None):
        ''' called by aws subcribe '''
        LOGGER.info('got broadcast: %s', message)
        args = json_utils.loads(message)
        asyncio.ensure_future(self.consumer(*args))



"""
    A service is a task that runs in the background.
    It can be started and stopped.
"""
import asyncio
import logging

LOGGER = logging.getLogger(__name__)


class ServicesManager:

    _SERVICE_ACTIONS_ = [
        "start_service",
        "stop_service",
        "start_all_services",
        "stop_all_services",
        "available_services",
        "status_services"
    ]

    def __init__(self, services):
        self._services = {srv.__name__: srv for srv in self._discover_(services)}
        self._tasks = {}

    @classmethod
    def _discover_(cls, services):
        """ utility to discover public functions in a module """
        _services = services.__all__ if hasattr(services, "__all__") else dir(services)
        for name in _services:
            if name[0] == "_":
                continue
            service = getattr(services, name)
            if callable(service):
                yield service

    @property
    def available_services(self):
        """ names of available services """
        return [name for name in self._services]

    async def status_services(self):
        """ returns a list of services and a list of running """
        return {
            "running": [str(key) for key in self._tasks],
            "services": self.available_services
        }

    async def start_service(self, name: str):
        """ create a task of a function """
        if name not in self._tasks:
            LOGGER.info("running service %s", name)
            task = asyncio.ensure_future(self._services[name]())
            self._tasks[name] = task

    async def stop_service(self, name: str):
        """ stop a task """
        if name in self._tasks:
            LOGGER.info("stopping service %s", name)
            task = self._tasks[name]
            task.cancel()
            LOGGER.info("stopped service %s", name)
            del self._tasks[name]

    async def start_all_services(self):
        """ run all services """
        for name in self.available_services:
            await self.start_service(name)

    async def stop_all_services(self):
        """ helper to shut down """
        for name in list(self._tasks):
            await self.stop_service(name)



""" Utilities for working with redis """
import asyncio
import logging
import aioredis
from bluemax.settings import get_settings

LOGGER = logging.getLogger(__name__)


async def connect(redis_url: str = None):
    """
        Create a single connection to redis using
        option.redis_url if none is supplied
    """
    redis_url = get_settings("redis_url") if redis_url is None else redis_url
    assert redis_url
    if redis_url.startswith("redis://:"):
        LOGGER.info("connecting to redis: redis://%s", redis_url.split('@')[-1])
    else:
        LOGGER.info("connecting to redis: %s", redis_url)
    loop = asyncio.get_event_loop()
    redis_conn = await aioredis.create_redis(redis_url, loop=loop)
    return redis_conn


_POOLS_ = {}


async def pool(redis_url: str = None):
    """
        Create a pool connection to redis using
        option.redis_url if none is supplied
    """
    redis_url = get_settings("redis_url") if redis_url is None else redis_url
    assert redis_url
    if redis_url.startswith("redis://:"):
        LOGGER.info("connecting to redis: redis://%s", redis_url.split('@')[-1])
    else:
        LOGGER.info("connecting to redis: %s", redis_url)
    con_pool = _POOLS_.get(redis_url)
    if con_pool is None:
        loop = asyncio.get_event_loop()
        con_pool = await aioredis.create_pool(
            redis_url, minsize=5, maxsize=10, loop=loop
        )
        _POOLS_[redis_url] = con_pool
    return con_pool


async def shutdown():
    for url in list(_POOLS_):
        redis_pool = _POOLS_[url]
        redis_pool.close()
        await redis_pool.wait_closed()
        del _POOLS_[url]
    _POOLS_.clear()



"""
    We need to use alembic to host multiple projects in one directory
"""
import os
import shutil
from configparser import SafeConfigParser
from alembic.config import Config
from invoke import task
from bluemax.tasks.utils import confirm_action, Colored


def get_db_url(name):
    """ get the db_url from alembic.ini """
    alembic_cfg = Config("alembic.ini")
    return alembic_cfg.get_section_option(name, "sqlalchemy.url")


def confirm_db_action(action, name):
    """db are you sure? """
    db_url = get_db_url(name)
    return confirm_action(f"{action} {db_url}")


def _run_with_pythonpath(ctx, cmd):
    """ extend the python path environment variable with cwd """
    path = os.getenv("PYTHONPATH", "")
    env = {"PYTHONPATH": ".:" + path}
    return ctx.run(cmd, env=env)


def revise_ini(adding=None, removing=None):
    """ add or remove from ini file """
    parser = SafeConfigParser()
    with open("alembic.ini", "r") as configfile:
        parser.read_file(configfile)

    if parser.has_section("alembic"):
        del parser["alembic"]
    if adding:
        parser.add_section(adding)
        parser.set(adding, "script_location", f"database/{adding}")
        parser.set(adding, "sqlalchemy.url", f"sqlite:///{adding}.sqlite3")
    elif parser.has_section(removing):
        del parser[removing]

    with open("alembic.ini", "w") as configfile:
        parser.write(configfile)


@task(help={"project": "name of the project to add a model package to"})
def create_model(_, project):
    """ creates a sqlalchemy model package inside project """
    model_path = os.path.join(project, "model")
    if os.path.isdir(model_path):
        print("model exists")
        return
    os.makedirs(model_path)
    if not os.path.isfile(os.path.join(project, "__init__.py")):
        open(os.path.join(project, "__init__.py"), "w")
    open(os.path.join(model_path, "__init__.py"), "w")
    with open(os.path.join(model_path, "base.py"), "w") as file:
        file.write(
            '''""" Used by alembic """
from sqlalchemy.ext.declarative import declarative_base, declared_attr


class Base:
    """ Base for orm model classes """

    @declared_attr
    def __tablename__(cls):
        """ with this tablename is class name lower"""
        return cls.__name__.lower()


Base = declarative_base(cls=Base)
'''
        )


@task(
    help={
        "name": "name of the database",
        "project": "where model is created, defaults to name of database",
    }
)
def create(ctx, name, project=None):
    """ creates an alembic managed db """
    if project is None:
        project = name
    if not os.path.isdir("database"):
        os.makedirs("database")

    ctx.run(f"alembic init database/{name}")
    revise_ini(name)

    with open(os.path.join("database", name, "env.py"), "r") as file:
        contents = file.read()
        contents = contents.split("target_metadata = None\n")

    with open(os.path.join("database", name, "env.py"), "w") as file:
        file.write(contents[0])
        file.write("\n")
        file.write(f"from {project}.model.base import Base\n")
        file.write("target_metadata = Base.metadata\n")
        file.write(contents[1])

    create_model(ctx, project)


@task(help={"name": "name of the database"})
def remove(_, name):
    """ removes a database of name from the database folder """
    db_path = os.path.abspath(os.path.join("database", name))
    if os.path.isdir(db_path):
        shutil.rmtree(db_path)
        print(f"removed {db_path}")
    revise_ini(removing=name)


@task(help={"name": "name of the database", "message": "revision title"})
def revise(ctx, name, message):
    """ creates an autogenerated revision in alembic """
    _run_with_pythonpath(
        ctx, f'alembic -n {name} revision --autogenerate -m "{message}"'
    )


@task(
    help={"name": "name of the database", "revision": "almebic revision to upgrade to"}
)
def upgrade(ctx, name, revision="head"):
    """ upgrades with alembic to revision """
    if confirm_db_action(Colored.green("UPGRADE"), name):
        _run_with_pythonpath(ctx, f"alembic -n {name} upgrade {revision}")


@task(
    help={
        "name": "name of the database",
        "revision": "almebic revision to downgrade to",
    }
)
def downgrade(ctx, name, revision="base"):
    """ downgrades to revsion with alembic - base is nothing!"""
    if confirm_db_action(Colored.red("DOWNGRADE"), name):
        _run_with_pythonpath(ctx, f"alembic -n {name} downgrade {revision}")



"""
    This context is available either remote or local
    It is maintained by the manager.
"""
__all__ = ["current_user", "broadcast", "broadcast_on_success"]

import logging
import contextvars
import contextlib

LOGGER = logging.getLogger(__name__)

_USER_ = contextvars.ContextVar("_USER_")
_BROADCAST_ON_SUCCESS_ = contextvars.ContextVar("_BROADCAST_ON_SUCCESS_")

# one to rule them all
_MANAGER_ = None


def current_user():
    """ returns the user(actor) for this procedure call """
    return _USER_.get(None)


def broadcast(signal, message, filter_clients=None):
    """ This allows for the broadcast to client inline """
    _MANAGER_.broadcast(signal, message, filter_clients)


def broadcast_on_success(signal, message, filter_clients=None):
    """ This allows for the boradcast to client after func return """
    _BROADCAST_ON_SUCCESS_.get([]).append((signal, message, filter_clients))


def perform(qname, *args, **kwargs):
    """ Returns an awaitable of the result of calling qname """
    LOGGER.debug("calling manager %s with %s", _MANAGER_, qname)
    return _MANAGER_.perform(current_user(), qname, *args, **kwargs)


@contextlib.contextmanager
def user_call(user=None):
    """ With this we setup contextvars and reset """
    utoken = _USER_.set(user)
    btoken = _BROADCAST_ON_SUCCESS_.set([])
    try:
        yield
    finally:
        _USER_.reset(utoken)
        _BROADCAST_ON_SUCCESS_.reset(btoken)


def context_call(func, user=None, args=None, kwargs=None):
    """ Setup context and call """
    with user_call(user):
        result = func(*args, **kwargs)
        broadcasts = _BROADCAST_ON_SUCCESS_.get()
        if broadcasts:
            _MANAGER_.broadcast_later(broadcasts)
        return result


async def acontext_call(func, user=None, args=None, kwargs=None):
    """ Setup context and call async """
    with user_call(user):
        result = await func(*args, **kwargs)
        broadcasts = _BROADCAST_ON_SUCCESS_.get()
        if broadcasts:
            _MANAGER_.broadcast_later(broadcasts)
        return result



"""
    MS Auth
"""
import logging
from urllib.parse import urlencode
import tornado.web
import tornado.auth
from bluemax.utils import json_utils
from .login_handler import LoginHandler

MS_OAUTH2 = "https://login.microsoftonline.com/common/oauth2"


class MSLoginHandler(LoginHandler, tornado.auth.OAuth2Mixin):  # pylint: disable=W0223
    """ Implementation of Azure authentication """

    _USER_URL = "https://graph.microsoft.com/v1.0/me"
    _OAUTH_AUTHORIZE_URL = f"{MS_OAUTH2}/v2.0/authorize"
    _OAUTH_ACCESS_TOKEN_URL = f"{MS_OAUTH2}/v2.0/token"

    async def get(self):
        """ respond to first and redirected call """
        if self.get_argument("code", False):
            access = await self.get_authenticated_user(
                redirect_uri=self.settings["login_url"],
                client_id=self.settings["ms_oauth"]["key"],
                client_secret=self.settings["ms_oauth"]["secret"],
                code=self.get_argument("code", None),
                extra_params={"scope": "User.Read profile openid email"},
            )
            user = await self.get_user_profile(
                self._USER_URL, access_token=access["access_token"]
            )
            logging.info(user)
            self.set_current_user(user)
            self.redirect(self.get_argument("state", "/"))
        else:
            await self.authorize_redirect(
                redirect_uri=self.settings["login_url"],
                client_id=self.settings["ms_oauth"]["key"],
                client_secret=self.settings["ms_oauth"]["secret"],
                scope=["User.Read"],
                response_type="code",
                extra_params={
                    "approval_prompt": "auto",
                    "state": self.get_argument("next", "/"),
                },
            )

    async def get_user_profile(self, url, access_token):
        """ asks microsof for the user details """
        http = self.get_auth_http_client()
        try:
            response = await http.fetch(url, headers={"Authorization": access_token})
        except Exception as ex:
            logging.exception(ex)
            raise
        logging.debug(response.body)
        return json_utils.loads(response.body)

    async def get_authenticated_user(
        self, redirect_uri, client_id, client_secret, code, extra_params=None
    ):
        """ logs in """
        http = self.get_auth_http_client()
        args = {
            "redirect_uri": redirect_uri,
            "code": code,
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "authorization_code",
        }
        if extra_params:
            args.update(extra_params)
        body = urlencode(args)
        try:
            response = await http.fetch(
                self._OAUTH_ACCESS_TOKEN_URL,
                method="POST",
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                body=body,
            )
        except Exception as ex:
            logging.info(body)
            logging.exception(ex)
            raise
        logging.debug(response.body)
        args = json_utils.loads(response.body)
        session = {
            "access_token": args.get("access_token"),
            "expires_in": args.get("expires_in"),
        }
        return session



"""
    Implements broadcasting to redis
"""
import asyncio
import logging
from bluemax import get_settings
from bluemax.web.broadcaster import Broadcaster
from bluemax.utils import json_utils
from .redis_pubsub import redis_publish, redis_subscribe

LOGGER = logging.getLogger(__name__)


class RedisBroadcaster(Broadcaster):
    """
    Connects the local queue to redis and
    redis to the local broadcaster(consumer)
    """

    def __init__(self):
        self.consumer = None
        super().__init__()

    async def broadcaster(self, queue, consumer):
        """ Connect websockets to manager broadcast queue """
        self.consumer = consumer
        LOGGER.info("subscribing to redis")
        self.handle_broadcast()
        LOGGER.info("listening to queue")
        while True:
            value = await queue.get()
            self._publish_(*value)
            queue.task_done()

    @redis_publish(key=get_settings("redis_broadcast_q", "broadcast"))
    def _publish_(self, signal, message, client_filter=None):
        """ will broadcast to redis """
        msg = json_utils.dumps((signal, message, client_filter))
        LOGGER.debug("broadcasted to redis: %s", msg)
        return msg

    @redis_subscribe(key=get_settings("redis_broadcast_q", "broadcast"))
    async def handle_broadcast(self, message=None):
        """ called by redis sub """
        LOGGER.info("got broadcast: %s", message)
        args = json_utils.loads(message)
        asyncio.ensure_future(self.consumer(*args))



"""
    Introspective helpers for serializing SqlAlchemy Objects
"""
import logging
from sqlalchemy.orm import class_mapper, ColumnProperty

LOGGER = logging.getLogger(__name__)


def dumps(item):
    """ returns a dictionary of item properties """
    result = {"_type_": item.__class__.__name__}
    for prop in class_mapper(item.__class__).iterate_properties:
        if prop.key[0] == "_":
            continue
        if isinstance(prop, ColumnProperty):
            result[prop.key] = getattr(item, prop.key)
    return result


def loads(values, item):
    """ sets properties from values"""
    for prop in class_mapper(item.__class__).iterate_properties:
        if prop.key[0] == "_":
            continue
        if isinstance(prop, ColumnProperty) and prop.key in values.keys():
            setattr(item, prop.key, values.get(prop.key))



"""
    a description of a procedure
"""
from collections import OrderedDict
import inspect
import logging

LOGGER = logging.getLogger(__name__)


def describe(target):
    """
        Will inspect each function in target.__all__
        and return an instance of Procedure for each
    """
    result = OrderedDict()
    for key in target.__all__:
        f = getattr(target, key)
        s = Procedure(key, f)
        LOGGER.debug(s)
        result[key] = s
    return result


class Procedure:
    """
       a description of a procedure
    """

    def __init__(self, key, f):
        self.name = key
        self.desc = inspect.signature(f)
        self.docs = inspect.getdoc(f)
        self.f = f

    def __repr__(self):
        return f"{self.name} {self.desc} - {self.docs}"

    def parse_http_kwargs(self, values):
        """
            type conversion for int and float args
        """
        params = self.desc.parameters
        for k, v in values.items():
            if k in params and v is not None:
                if params[k].annotation is int:
                    values[k] = int(v)
                if params[k].annotation is float:
                    values[k] = float(v)

    @classmethod
    def annotation_to_str(cls, annotation):
        if annotation is None:
            return None
        if isinstance(annotation, str):
            return annotation
        if hasattr(annotation, "__name__"):
            return annotation.__name__
        return annotation

    def to_json(self):
        """ so it can be dumped """
        return OrderedDict(
            [
                ("name", self.name),
                ("params", [str(p) for p in self.desc.parameters.values()]),
                (
                    "returns",
                    self.annotation_to_str(self.desc.return_annotation)
                    if self.desc.return_annotation is not self.desc.empty
                    else "",
                ),
                ("docs", self.docs),
                ("auth", hasattr(self.f, 'auth')),
            ]
        )




from importlib import import_module


def qname_to_class(qname):
    """ returns a class from module_name:class_name """
    if isinstance(qname, str):
        _modules, _class = qname.split(":")
        return getattr(import_module(_modules), _class)
    return qname



"""
    Worker listens to a queue in redis and performs
    work as an Rpc Manager
"""
import asyncio
import logging
from bluemax.settings import get_settings
from bluemax.rpc import ContextRpc
from bluemax.utils import json_utils
from bluemax.performer import Performer
from .redis_pubsub import redis_publish, redis_cmd
from . import rd_session


LOGGER = logging.getLogger(__name__)

"""
    Main worker file. It subscribes to the
    redis queue and broadcasts activity
"""


class Worker(ContextRpc, Performer):
    """ A context rpc fed by redis """

    def __init__(self, *args, **kwargs):
        """ setup variables before run """
        ContextRpc.__init__(self, *args, **kwargs)
        Performer.__init__(self)
        # set off subscription
        self.handle_work()

    async def shutdown(self, wait=True):
        await ContextRpc.shutdown(self)

    @redis_cmd(key=get_settings("redis_work_q", "perform"), cmd="blpop", cmd_args=[0])
    async def handle_work(self, val=None):
        """ unpack the mesage and call perform """
        LOGGER.info("got work %s", val)
        if val is None:
            return
        content = json_utils.loads(val[1].decode("utf-8"))
        try:
            LOGGER.info("%r", content["qname"])
            result = await self.perform(
                content["user"], content["qname"], *content["args"], **content["kwargs"]
            )
            content["result"] = result
        except asyncio.CancelledError:
            LOGGER.info("cancelled")
            return
        except Exception as ex:  # pylint: disable=W0703
            content["error"] = {"code": -32000, "message": str(ex)}
        if content.get("reply"):
            pool = await rd_session.pool()
            await pool.execute("publish", content["reply"], json_utils.dumps(content))
            LOGGER.info("published reply %s", content)

    @redis_publish(key=get_settings("redis_broadcast_q", "broadcast"))
    def _broadcast_(self, signal, message, client_filter=None):
        """ will broadcast to redis """
        msg = json_utils.dumps((signal, message, client_filter))
        LOGGER.debug("broadcasted to redis: %s", msg)
        return msg


def main():
    """ Creates Worker and starts tornado """
    assert get_settings("redis_url")
    assert get_settings("procedures")
    worker = Worker.rpc_for(
        get_settings("procedures"),
        max_workers=get_settings("max_workers"),
        max_threads=get_settings("max_threads"),
    )
    loop = asyncio.get_event_loop()
    try:
        loop.run_forever()
    finally:
        loop.run_until_complete(worker.shutdown())
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()



"""
    A broadacasting dictionary that waits a loop for changes
"""
import time
from bluemax.utils import json_utils


class Status:
    """
        Simple helper class of dictionary
        that calls you every time it changes
    """

    def __init__(self, ioloop, callback, values=None, timeout=0.1):
        self.ioloop = ioloop
        self._items_ = values if values is not None else {}
        self._last_ = None
        self._callback_ = callback
        self._timeout_ = timeout
        self._pending_ = None

    def __getitem__(self, key):
        """ have an item """
        return self._items_.get(key)

    def __setitem__(self, key, value):
        """ change and flag we need to test for change """
        self._items_[key] = value
        if self._pending_ is not None:
            self.ioloop.remove_timeout(self._pending_)
        self._pending_ = self.ioloop.add_timeout(
            time.time() + self._timeout_, self._do_callback_
        )

    def _do_callback_(self):
        """ if we've change call our callback """
        self._pending_ = None
        now = json_utils.dumps(self)
        if self._last_ != now:
            self._last_ = now
            self._callback_(self)

    @property
    def items(self):
        """ return the data we're holding """
        return self._items_

    def to_json(self):
        """ We work well with json_utils """
        return self._items_

    def __repr__(self):
        """ we print nicely """
        return repr(self._items_)



"""
    Proxy to procedures

    ...because it walks like a duck....

    should you wish to run locally::

        async with Moo('<package name>') as cow:
            result = await cow.add(2, 2)
            assert result == 4

"""

import logging
from bluemax import bring_up, get_settings
from bluemax.utils import qname_to_class

LOGGER = logging.getLogger(__name__)


class Moo:  # pylint: disable=R0903
    """ we want a proxy """

    def __init__(self, procedures, manager="bluemax.rpc:PoolRpc"):
        bring_up(procedures)
        rpc = qname_to_class(manager)
        self._rpc_ = rpc.rpc_for(get_settings("procedures"))

    def _stop_(self):
        """ call this to shutdown the rpc """
        return self._rpc_.shutdown()

    async def __aenter__(self):
        """ act as an async context manager """
        return self

    async def __aexit__(self, type_, value, traceback):
        """ act as an async context manager """
        await self._stop_()

    def __getattribute__(self, name):
        """ public attributes are proxies to procedures """
        LOGGER.debug("attribute %s", name)
        if name[0] == "_":
            return super().__getattribute__(name)
        if name in self._rpc_.procedures:

            def _wrapped_(*args, **kwargs):
                LOGGER.debug("calling %s(*%s, ***%s)", name, args, kwargs)
                return self._rpc_.perform(name, *args, **kwargs)

            return _wrapped_
        raise AttributeError(f"no such procedure: {name}")



'''
    We wanted something like Thread and Process Pool Exceutors
    but for async function
'''
import asyncio
import logging
from concurrent.futures import Executor

LOGGER = logging.getLogger(__name__)


class NotRunningException(Exception):
    ''' Called when a running pool is required '''


class AsyncPoolExecutor(Executor):
    '''
        This object can be called with async functions and you
        can limit the number of workers with max_workers.
        It maintains a queue of work of max_size.
    '''

    def __init__(self, loop=None, max_workers=None, max_size=0):
        self.max_workers = max_workers if max_workers else 4
        self.running = False
        self.loop = loop if loop else asyncio.get_event_loop()
        self.queue = asyncio.Queue(maxsize=max_size, loop=self.loop)
        self.workers = []

    def start(self):
        ''' called by aenter '''
        LOGGER.debug('starting...')
        self.running = True
        for n_id in range(self.max_workers):
            worker = asyncio.ensure_future(self.worker(n_id))
            self.workers.append(worker)

    async def worker(self, n_id):
        ''' worker called a task '''
        LOGGER.debug('working: %s', n_id)
        while self.running is True:
            try:
                await self.perform(n_id)
            except asyncio.CancelledError:
                return

    async def perform(self, n_id):
        ''' called by a worker to get queue item '''
        LOGGER.debug('pulling from queue: %s', n_id)
        item = await self.queue.get()
        results, func, args, kwargs = item
        try:
            result = await func(*args, **kwargs)
            results.set_result(result)
        except Exception as ex:  # pylint: disable=W0703
            results.set_exception(ex)
        finally:
            self.queue.task_done()

    def submit(self, fn, *args, **kwargs):  # pylint: disable=W0221
        ''' part of the executor api '''
        if self.running is False:
            raise NotRunningException()
        result = self.loop.create_future()
        LOGGER.debug('adding to queue...')
        self.queue.put_nowait((result, fn, args, kwargs))
        return result

    def map(self, func, *iterables):  # pylint: disable=W0221
        ''' part of the executor api '''
        if self.running is False:
            raise NotRunningException()
        results = []
        args = zip(*iterables)
        LOGGER.debug('args... %r', args)
        for line in args:
            results.append(self.submit(func, *line))
        return results

    async def shutdown(self, wait=True):
        '''
            part of the executor api
            called by aexit
        '''
        LOGGER.debug('shutting down...')
        if wait:
            LOGGER.debug('joining...')
            await self.queue.join()
        self.running = False
        try:
            while True:
                LOGGER.debug('cancelling queue contents...')
                item = self.queue.get_nowait()
                item[0].cancel()
                self.queue.task_done()
        except asyncio.QueueEmpty:
            pass
        LOGGER.debug('cancelling workers...')
        for task in self.workers:
            task.cancel()
        LOGGER.debug('gathering workers...')
        await asyncio.gather(*self.workers, return_exceptions=True)
        self.workers.clear()
        LOGGER.debug('shutdown')

    async def __aenter__(self):
        ''' This makes us an async contextmanager '''
        self.start()
        return self

    async def __aexit__(self, *args):
        ''' This is called when the with is up '''
        await self.shutdown()



"""
    Web server with rpc handler
"""
import logging
import tornado.web
from tornado.ioloop import IOLoop, PeriodicCallback
from tornado.web import StaticFileHandler
from bluemax.settings import get_settings, extend
from bluemax.utils import qname_to_class
from .broadcaster import Broadcaster
from .authentication import LogoutHandler
from .auth_static_file_handler import AuthStaticFileHandler
from .rpc_websocket import RpcWebsocket, RpcHandler

BROADCASTER = Broadcaster()
LOGGER = logging.getLogger(__name__)


def make_app(loop=None):
    """ Return a tornado app with settings and routes """
    global BROADCASTER  # pylint: disable=W0603
    settings = get_settings("tornado", {})
    assert get_settings("procedures"), "procedures required to remote!"
    default_manager = (
        "bluemax.services:ServicesRpc"
        if get_settings("services")
        else "bluemax.rpc:ContextRpc"
    )
    manager = qname_to_class(settings.get("manager", default_manager))
    settings["manager"] = manager.rpc_for(
        get_settings("procedures"), services=get_settings("services"), loop=loop
    )
    settings.setdefault("debug", get_settings("debug", False))
    routes = [(r"/rpc", RpcHandler), (r"/ws", RpcWebsocket)]
    if settings.get("auth_handler"):
        settings.setdefault("cookie_name", "bluemax-user")
        settings.setdefault("cookie_secret", "<change me - it was a dark and stormy..>")
        auth_handler = qname_to_class(settings.get("auth_handler"))
        settings["login_url"] = "/login"
        routes.insert(0, (r"/logout", LogoutHandler))
        routes.insert(0, (r"/login", auth_handler))
        if settings.get("static_dir"):
            # we use this because out static_dir is root for spa
            # static_path created a static by tornado
            routes.append(
                (
                    r"/(.*)",
                    AuthStaticFileHandler,
                    {
                        "path": settings.get("static_dir"),
                        "default_filename": "index.html",
                    },
                )
            )
            LOGGER.info("serving from %s", settings.get("static_dir"))
    elif settings.get("static_dir"):
        # we use this because out static_dir is root for spa
        # static_path created a static by tornado
        routes.append(
            (
                r"/(.*)",
                StaticFileHandler,
                {"path": settings.get("static_dir"), "default_filename": "index.html"},
            )
        )
        LOGGER.info("serving from %s", settings.get("static_dir"))
    if get_settings("urls_extend"):
        routes = extend(get_settings("urls_extend"), routes)
    LOGGER.debug("tornado routes: %s", routes)
    LOGGER.debug("tornado settings: %s", settings)
    app = tornado.web.Application(routes, **settings)
    if get_settings("broadcaster"):
        BROADCASTER = qname_to_class(get_settings("broadcaster"))()
    BROADCASTER.start(settings["manager"].get_broadcast_queue(), RpcWebsocket.broadcast)
    return app


def main():
    """ run the tornado web server """
    app = make_app()
    port = get_settings("bluemax", {}).get("port", get_settings("PORT", "8080"))
    address = get_settings("bluemax", {}).get("address", "127.0.0.1")
    app.listen(int(port), address)
    LOGGER.info("listening on port %s", port)
    if app.settings.get("debug") is True:
        LOGGER.info("running in debug mode")
    ioloop = IOLoop.current()
    keep_alive_interval = int(get_settings("keep_alive_interval", "30000"))
    PeriodicCallback(RpcWebsocket.keep_alive, keep_alive_interval).start()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        LOGGER.info("stopping")
        ioloop.stop()



"""
    A json rpc 2.0 websocket
"""
import asyncio
import logging
import time
import uuid
from tornado.web import RequestHandler, HTTPError
from tornado.websocket import WebSocketHandler
from bluemax.utils import json_utils
from .authentication import UserMixin

LOGGER = logging.getLogger(__name__)


class JsonRpcException(Exception):
    """ Throw xception that capture message """

    def __init__(self, code, message):
        super().__init__(self, message)
        self.js_code = code
        self.js_message = message


class JsonRpcMixin:
    """ Common functions used by both ws and http """

    @property
    def manager(self):
        """ return the manager setting """
        return self.settings["manager"]

    @property
    def procedures(self):
        """ returns the managers procedure names """
        return self.manager.procedures

    def is_authenticated(self):
        """
        Raise an error if we should login otherwise
        return user
        """
        user = None
        if self.settings.get("login_url", None):
            user = self.get_current_user()
            if user is None:
                raise HTTPError(403)
        return user

    @classmethod
    def get_params(cls, content):
        """ split params into args and kwargs """
        if isinstance(content["params"], dict):
            return [], content["params"]
        if isinstance(content["params"], list):
            return content["params"], {}
        raise JsonRpcException(-32602, "Params neither list or dict")

    async def handle_content(self, content):  # noqa: 901
        """ Assumes a json 2.0 method conent """
        LOGGER.info(content)
        result = {"jsonrpc": "2.0"}
        try:
            if content.get("jsonrpc") != "2.0":
                raise JsonRpcException(-32600, "protocol not supported")
            ref = content.get("id", None)
            if ref:
                result["id"] = ref

            method = content.get("method")
            if method is None:
                raise JsonRpcException(-32600, "no method")
            if method[0] == "_":
                raise JsonRpcException(-32600, "method private")
            if method not in self.procedures:
                raise JsonRpcException(-32600, f"no such method: {method}")

            user = self.is_authenticated()
            args, kwargs = self.get_params(content)
            logging.debug("%s %s %s %s", user, method, args, kwargs)
            response = self.manager.perform(user, method, *args, **kwargs)
            if response:  # can return None or future
                response = await response
            if asyncio.iscoroutine(response) or asyncio.isfuture(response):
                result["result"] = await response
            else:
                result["result"] = response
            if ref is None:
                result = None  # its a notification

        except JsonRpcException as ex:
            LOGGER.exception(ex)
            result["error"] = {"code": ex.js_code, "message": ex.js_message}
        except Exception as ex:  # pylint: disable=W0703
            LOGGER.exception(ex)
            result["error"] = {"code": -32000, "message": str(ex)}
        LOGGER.debug(repr(result))
        return result


class RpcHandler(UserMixin, JsonRpcMixin, RequestHandler):  # pylint: disable=W0223
    """ marshall json rpc 2.0 requests and respond """

    async def post(self):
        """ We only hanlde the post method """
        self.is_authenticated()
        content = json_utils.loads(self.request.body)
        result = await self.handle_content(content)
        if result:
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            self.write(json_utils.dumps(result))


class RpcWebsocket(UserMixin, JsonRpcMixin, WebSocketHandler):  # pylint: disable=W0223
    """ implementation """

    _clients_ = []

    def filter_client(self, filter_clients=None):  # pylint: disable=W0613,R0201
        """ Does this client meet these criteria """
        return False

    @classmethod
    def keep_alive(cls):
        msg = str(time.time()).encode("utf8")
        for client in cls._clients_:
            client.ping(msg)

    @classmethod
    async def broadcast(cls, signal, message, filter_clients=None):
        """ broadcasts signal:message to all clients """
        LOGGER.debug("broadcasting %s: %s", signal, message)
        data = json_utils.dumps({"signal": signal, "message": message})
        for client in cls._clients_:
            if not client.filter_client(filter_clients):
                client.write_message(data)

    def initialize(self):
        """ set our id """
        self._id = str(uuid.uuid4())  # pylint: disable=W0201

    def open(self, *args, **kwargs):
        """ websocket open - register user """
        user = self.is_authenticated()
        self._clients_.append(self)
        if user:
            LOGGER.info("user: %s", user)
            self.write_message(
                json_utils.dumps({"signal": "set_user", "message": user})
            )

    async def on_message(self, data):  # pylint: disable=W0221
        """ handle the action """
        content = json_utils.loads(data)
        loop = asyncio.get_running_loop()
        asyncio.ensure_future(self.handle_result(content), loop=loop)

    async def handle_result(self, content):
        """ handle the result of a perform """
        result = await self.handle_content(content)
        if result:
            self.write_message(json_utils.dumps(result))

    def on_close(self):
        """ unregister us """
        self._clients_.remove(self)



"""
    Helpers for SqlAlchemy connections
"""
import logging
from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from bluemax.settings import get_settings

LOGGER = logging.getLogger(__name__)


class ConnectionCache:
    """ Cache of Engine and Session for db_urls """

    connections = {}

    def __init__(self, name, url, **kwargs):
        self.name = name
        self.url = url
        LOGGER.debug("connecting to db: %s", url)
        self.session, self.engine = self.connect(url, **kwargs)
        self.connections[self.name] = self

    @classmethod
    def get(cls, name):
        """ return cached connection """
        return cls.connections.get(name)

    @classmethod
    def connect(cls, db_url, echo=False, **kwargs):
        """ Connect to db and provide engine and Session """
        assert db_url, "db_url required"
        engine_ = create_engine(db_url, echo=echo, **kwargs)
        session_ = sessionmaker(bind=engine_)
        return session_, engine_

    @classmethod
    def setup_db(cls, base_name="db_url", db_url=None, **kwargs):
        """ Create a connection and cache it """
        logging.debug("setUp(%r)", base_name)
        session_ = cls.get(base_name)
        if session_ is None:
            url = db_url if db_url else get_settings(base_name)
            session_ = cls(base_name, url, **kwargs)
        return session_.engine, session_.session


def setup_db(cls, base_name="db_url", db_url=None, **kwargs):
    # for ease of use - proxy to ConnectionCache
    return ConnectionCache.setup_db(base_name=base_name, db_url=db_url, **kwargs)


@contextmanager
def session(session_cls):
    """Provide a transactional scope around a series of operations."""
    session_obj = session_cls()
    try:
        yield session_obj
        session_obj.commit()
    except Exception:
        session_obj.rollback()
        raise
    finally:
        session_obj.close()


def session_scope(base_name="default", db_url=None):
    """Provide a transactional scope around a series of operations."""
    if ConnectionCache.get(base_name) is None:
        ConnectionCache.setup_db(base_name, db_url)
    return session(ConnectionCache.get(base_name).session)



"""
    We want to encode datetimes and Decimals to json
"""
import json
from decimal import Decimal
from collections.abc import Callable


class DateTimeEncoder(json.JSONEncoder):
    """
        Encodes datetimes and Decimals
        calls to_json on object if it has that method
    """

    def default(self, obj):  # pylint: disable=W0221,R1710,E0202
        """ return json types """
        if hasattr(obj, "isoformat"):
            return obj.isoformat().replace("T", " ")
        if isinstance(obj, Decimal):
            return float(obj)
        if hasattr(obj, "to_json") and isinstance(getattr(obj, "to_json"), Callable):
            return obj.to_json()


def loads(*args, **kwargs):
    """ calls json.loads """
    return json.loads(*args, **kwargs)


def dumps(obj, **kwargs):
    """ calls json.loads with DateTimeEncoder """
    return json.dumps(obj, cls=DateTimeEncoder, **kwargs)



from invoke import task
from . import create


@task
def sidney(ctx, name):
    ''' all the jazz - creates everything.'''
    from bluemax.sa import tasks as database

    create.project(ctx, name, with_config=True)
    create.tests(ctx, name)
    create.docker(ctx, name)
    database.create(ctx, name)



"""
    A service is a task that runs in the background.
    It can be started and stopped.
"""
import asyncio
import importlib
import logging
from bluemax import context
from bluemax.settings import get_settings
from bluemax.work.redis_pubsub import redis_subscribe
from bluemax.utils import json_utils
from bluemax.performer import Performer
from .services_manager import ServicesManager

LOGGER = logging.getLogger(__name__)


class Services(ServicesManager, Performer):
    """ This service manager only support async functions """

    def __init__(self, services):
        ServicesManager.__init__(self, services)
        Performer.__init__(self)
        self.handle_action()  # setup subscription
        LOGGER.debug("services %s", self.available_services)

    @redis_subscribe(key=get_settings("redis_services_q", "service"))
    async def handle_action(self, message=None):  # pylint: disable=R0201
        """ start and stop stuff """
        LOGGER.info(message)
        content = json_utils.loads(message)
        qname = content["qname"]
        if qname in self._SERVICE_ACTIONS_:
            func = getattr(self, qname)
            if asyncio.iscoroutinefunction(func):
                await func(*content["args"], **content["kwargs"])
            else:
                func(*content["args"], **content["kwargs"])

    async def shutdown(self):
        """ tidy up while we're here """
        await self.stop_all_services()
        LOGGER.info("shutting down redis subscriptions")


def main():
    """ run loop """
    assert get_settings("services")
    LOGGER.info("loading: %s", get_settings("services"))
    services_package = importlib.import_module(get_settings("services"))
    manager = Services(services_package)
    context._MANAGER_ = manager  # pylint: disable=W0212
    LOGGER.info("services: %s", manager.available_services)
    loop = asyncio.get_event_loop()
    asyncio.ensure_future(manager.start_all_services())
    try:
        loop.run_forever()
    finally:
        loop.run_until_complete(manager.shutdown())
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()



"""
    An rpc concept for asyncio
"""
from .asyncpoolexecutor import AsyncPoolExecutor, NotRunningException
from .tasks.run import bring_up
from .rpc import ContextRpc
from .settings import get_settings
from .main import Bluemax

VERSION = "0.2.6"
name = "bluemax"



"""
    Implementation of tornado authentication
"""
from .user_mixin import UserMixin
from .login_handler import LoginHandler
from .logout_handler import LogoutHandler
from .profile_handler import ProfileHandler



""" root level task - not imported with module """
import os
import signal
import logging
from invoke import task

LOGGER = logging.getLogger(__name__)


@task(help={
    "process": "server, worker or services"
})
def stop(ctx, process):
    """ stop a running server, worker or services """
    pid_file = f"{process}.pid"
    try:
        with open(pid_file, 'r') as f:
            contents = f.read()
        pid = int(contents)
        os.kill(pid, signal.SIGINT)
        os.remove(pid_file)
    except Exception:  # pylint: disable=W0703
        LOGGER.exception('problem with %s', pid_file)



"""
    Common functions used by Click and Invoke
"""
import os
import logging
import logging.config
import importlib
from contextlib import contextmanager
from bluemax.settings import extend

LOGGER = logging.getLogger(__name__)


class Colored:
    """ sometimes you need to colour console output """

    @staticmethod
    def green(text):
        """ turns text green """
        return f"\033[92m{text}\033[0m"

    @staticmethod
    def red(text):
        """ turns text red """
        return f"\033[91m{text}\033[0m"

    @staticmethod
    def blue(text):
        """ turns text blue """
        return f"\033[94m{text}\033[0m"

    @staticmethod
    def cyan(text):
        """ turns text cyan """
        return f"\033[36m{text}\033[0m"

    @staticmethod
    def gray(text):
        """ turns text gray """
        return f"\033[90m{text}\033[0m"

    @staticmethod
    def pink(text):
        """ turns text pink """
        return f"\033[33m{text}\033[0m"


@contextmanager
def pid_file(path):
    """ creates a file with the process id and removes it """
    pid = os.getpid()
    with open(path, "w") as file:
        file.write(f"{pid}")
    try:
        yield
    finally:
        if os.path.isfile(path):
            os.unlink(path)


def import_module(name):
    "Its ok not to have the module but not ok to not have dependants."
    try:
        logging.debug("looking for module %s", name)
        return importlib.import_module(name)
    except ModuleNotFoundError as ex:
        if name not in str(ex):
            raise


def get_module(name):
    """ import the project and return configuration options """
    options = {}
    if name is None:
        print("module expected")
    else:
        importlib.import_module(name)
        procs = importlib.import_module(f"{name}.procedures")
        if procs:
            if not getattr(procs, "__all__"):
                print(f"expected __all__ in procedures module")
            else:
                options["procedures"] = f"{name}.procedures"
                ext_log = import_module(f"{name}.log")
                if ext_log:
                    options["log_extend"] = f"{name}.log:extend"
                ext_settings = import_module(f"{name}.settings")
                if ext_settings:
                    options["settings_extend"] = f"{name}.settings:extend"
                ext_urls = import_module(f"{name}.urls")
                if ext_urls:
                    options["urls_extend"] = f"{name}.urls:extend"
        else:
            print(f"expected procedures in {name}")
    return options


def extend_logging(options):
    """ extend the logging and call dictConfig if response """
    if options.get("log_extend"):
        config = extend(options["log_extend"], {})
        if config:
            LOGGER.debug("logging.config %s", config)
            logging.config.dictConfig(config)


def confirm_action(action):
    """ Are you sure? """
    choice = ""
    while choice not in ["y", "n"]:
        choice = input(f"{action} [Y/N] ").lower()
    return choice == "y"



""" We need to stop this task """
import asyncio
import logging

LOGGER = logging.getLogger(__name__)


class Broadcaster:
    """ This broadcaster reads from a queue and passes to a consumer """

    def __init__(self):
        self.task = None
        self._broadcast_q_ = asyncio.Queue()

    def start(self, queue, consumer):
        """ start and store our task """
        self.task = asyncio.ensure_future(self.broadcaster(queue, consumer))

    async def stop(self):
        """ usung globals and module works """
        LOGGER.info("stopping broadcaster")
        if self.task:
            self.task.cancel()
            # await self.task

    async def broadcaster(self, queue, consumer):
        """ Connect websockets to manager broadcast queue """
        while True:
            value = await queue.get()
            await consumer(*value)
            queue.task_done()



# pylint: disable=W0621
"""
    Task for the running of bluemax
    The load order:
        if module:
            load module
            load settings
            extend setting
            load logging
            run target
        else
            load settings
            load module
            extend settings
            load logging
            run target
"""
import logging
from importlib import import_module
import yaml
from invoke import task
from bluemax import settings
from bluemax.utils.describe import describe
from .utils import get_module, pid_file, extend_logging, Colored

LOGGER = logging.getLogger(__name__)


def bring_up(module=None, config=None):
    """ the standard bring up order """
    if settings.SETTINGS:
        # only bring-up once
        return
    if config:
        settings.init_settings(config)
    if module is None:
        module = settings.SETTINGS["procedures"]
    assert module, "module required"
    mod = get_module(module)
    if mod:
        settings.SETTINGS.update(mod)
    extend_settings = settings.get_settings("settings_extend")
    if extend_settings:
        settings.extend(extend_settings, settings.SETTINGS)
    extend_logging(settings.SETTINGS)


@task
def config(_, module=None, config=None, services=None):
    """ displays resultant config """
    bring_up(module, config)
    LOGGER.info("\n---\n%s", yaml.dump(settings.SETTINGS))
    mod = import_module(settings.get_settings("procedures"))
    print(Colored.blue("available:"))
    for description in describe(mod).values():
        print("  ", description)
    services = settings.get_settings("services", services)
    if services:
        services = import_module(services)
        print(Colored.blue("services:"))
        for description in describe(services).values():
            print("  ", description)
    print("", flush=True)


@task(
    default=True,
    help={
        "pid": "/path/to/pid/file",
        "module": "module to remote",
        "services": "services module",
        "config": "path to config.yml",
    },
)
def server(_, module=None, pid="server.pid", services=None, config=None):
    """ runs a bluemax server with optional services """
    bring_up(module, config)
    if services:
        settings.SETTINGS["services"] = services

    from bluemax.web import server as m_server
    with pid_file(pid):
        m_server.main()


@task(
    help={
        "pid": "/path/to/pid/file",
        "module": "module to remote",
        "services": "services module",
        "config": "path to config.yml",
    }
)
def worker(_, module=None, pid="worker.pid", config=None):
    """ runs a bluemax worker """
    bring_up(module, config)

    from bluemax.work import worker as m_worker
    with pid_file(pid):
        m_worker.main()


@task(
    help={
        "pid": "/path/to/pid/file",
        "module": "services module",
        "config": "path to config.yml",
    }
)
def services(_, module=None, pid="services.pid", services=None, config=None):
    """ runs a bluemax services """
    bring_up(module, config)

    from bluemax.services import server as s_worker
    if services:
        settings.SETTINGS["services"] = services
    with pid_file(pid):
        s_worker.main()




'''
    Tools to create a project
'''
import os
import logging
from pathlib import Path
from invoke import task
from .utils import Colored

LOGGER = logging.getLogger(__name__)


def get_venv():
    ''' return the current virtual environment local path '''
    result = '.'
    cwd = os.getcwd()
    venv = os.getenv('VIRTUAL_ENV')
    LOGGER.debug('cwd: %s', cwd)
    LOGGER.debug('venv: %s', venv)
    if venv != cwd and venv.startswith(cwd):
        # remove cwd and trailing sep
        result = venv[len(cwd) + 1 :]
    return result


@task(default=True)
def project(_, name, force=False, with_config=False):
    ''' Creates a project with procedures, services, settings and urls. '''
    print(f'Creates project: {Colored.blue(name)}')
    if os.path.exists(name) and force is False:
        print('that module already exists: ', Colored.pink(name))
    else:
        if not os.path.isdir(name):
            os.makedirs(name)
        with open(os.path.join(name, '__init__.py'), 'w') as file:
            file.write(
                '''""" A bluemax project """
VERSION = "0.0.1"
'''
            )
        with open(os.path.join(name, 'settings.py'), 'w') as file:
            if with_config is True:
                file.write(
                    '''""" add your settings here """
import os
import logging
import yaml
from bluemax.settings import merge_dict


def extend(settings):
    """ change settings """
    logging.info("extending settings")
    stage = os.getenv("STAGE", "dev")
    stage_file = f"conf/{stage}.yml"
    base_file = "conf/base.yml"
    if os.path.isfile(base_file):
        with open(base_file, "r") as file:
            loaded = yaml.safe_load(file)
            if loaded:
                merge_dict(settings, loaded)
            logging.info("loaded config: base")
    if os.path.isfile(stage_file):
        with open(stage_file, "r") as file:
            loaded = yaml.safe_load(file)
            if loaded:
                merge_dict(settings, loaded)
            logging.info("loaded config: %s ", stage)
    return settings
'''
                )
            else:
                file.write(
                    '''""" add your settings here """
import logging


def extend(settings):
    """ change settings """
    logging.info('extending settings')
    return settings
'''
                )
        with open(os.path.join(name, 'urls.py'), 'w') as file:
            file.write(
                '''""" add your tornado routes here """
import logging


def extend(urls):
    """ add tornado routes to urls """
    logging.info('extending urls')
    return urls
'''
            )
        with open(os.path.join(name, 'procedures.py'), 'w') as file:
            file.write(
                '''""" an example function and exposing it through __all__ """

__all__ = ['add']


def add(int_a: int, int_b: int) -> int:
    """ simple addition of two integers"""
    return int_a + int_b
'''
            )
        with open(os.path.join(name, 'services.py'), 'w') as file:
            file.write(
                '''""" an example service """
import asyncio
import time
from bluemax import context

__all__ = ["clock"]


async def clock():
    """ Will broadcast every 5 seconds """
    while True:
        context.broadcast('time', {'now': time.time()})
        await asyncio.sleep(5)
'''
            )
        with open(os.path.join(name, 'log.py'), 'w') as file:
            file.write(
                '''""" an example log setup """
import os
import yaml


def extend(config):  # pylint: disable=W0613
    """
    config is the existing logging configuration
    returning a dict will be used with dictConfig
    """
    if os.path.isfile("logging.yml"):
        with open("logging.yml", "r") as file:
            config = yaml.safe_load(file)
    return config
'''
            )
        if not os.path.isfile('Makefile'):
            virtual_env = get_venv()
            with open(os.path.join('Makefile'), 'w') as file:
                file.write(
                    f'''
SHELL := /bin/bash

lint:
\tsource {virtual_env}/bin/activate && flake8 .
\tsource {virtual_env}/bin/activate && pylint {name} tests

test:
\tsource {virtual_env}/bin/activate && \\
\tpy.test --cov {name} --cov-report term-missing tests

check: lint test

run:
\tsource {virtual_env}/bin/activate && \\
\t\tbluemax run.server -m {name}

build:
\tsource {virtual_env}/bin/activate && \\
\t\tpython setup.py sdist

full: check build

setup:
\tif which python3.7 && [ ! -d {virtual_env} ] ; then python3.7 -m venv {virtual_env} ; fi
\tsource {virtual_env}/bin/activate \\
\t\t&& python -m pip install -q -U pip \\
\t\t&& pip install -e .[dev]
'''
                )
    if with_config is True:
        config(_, name, force=force)


@task
def tests(_, name, force=False):
    ''' create a config folder with dev, staging & prod yamls '''
    print(f'gen test: {name}')
    test_dir = 'tests'
    if os.path.exists(test_dir) and force is False:
        print('that folder already exists:', test_dir)
    else:
        if not os.path.isdir(test_dir):
            os.makedirs(test_dir)
        Path(os.path.join(test_dir, '__init__.py')).touch()
        with open(os.path.join(test_dir, 'conftest.py'), 'w') as file:
            file.write(
                f'''
import pytest
from bluemax.web.server import make_app
from bluemax.settings import SETTINGS

@pytest.fixture
def app(io_loop):
    SETTINGS.update(
        {{
            "procedures": "{name}.procedures",
            "tornado": {{
                "auth_handler": None,
                "manager": "bluemax:ContextRpc",
                "static_dir": None,
            }},
            "redis_url": None,
        }}
    )
    result = make_app()
    return result
'''
            )
        with open(os.path.join(test_dir, f'test_web.py'), 'w') as file:
            file.write(
                '''
import logging
from urllib.parse import urlencode
from bluemax.web.server import BROADCASTER
from bluemax.web import json_utils
from bluemax import context

async def test_add(http_server_client):
    message = {"jsonrpc": "2.0", "id": "1a", "method": "add", "params": [2, 2]}
    response = await http_server_client.fetch(
        "/rpc", method="POST", body=json_utils.dumps(message)
    )
    assert json_utils.loads(response.body)["result"] == 4

    await context._MANAGER_.shutdown()
    await BROADCASTER.stop()
'''
            )


@task
def config(_, name, conf='conf', force=False):
    ''' create a config folder with dev, staging & prod yamls '''
    print(f'gen config: {conf}')
    if os.path.exists(conf) and force is False:
        print('that folder already exists: ', conf)
    else:
        if not os.path.isdir(conf):
            os.makedirs(conf)
        with open(os.path.join(conf, 'base.yml'), 'w') as file:
            file.write(
                '''---
DEBUG: false
'''
            )
        with open(os.path.join(conf, 'dev.yml'), 'w') as file:
            file.write(
                f'''---
debug: true
procedures: {name}.procedures
services: {name}.services
log_extend: null
settings_extend: null
urls_extend: null
tornado:
    auth_handler: null
    manager: bluemax.services:servicesrpc
    static_dir: null
    cookie_name: null
    cookie_secret: null
    login_url: null
    cognito_oauth:
        key: null
        secret: null
        redirect_url: null
        endpoint: null
port: 8080
max_workers: 4
max_threads: 4
redis_url: none
redis_broadcast_q: broadcast
redis_work_q: perform
redis_services_q: services
'''
            )
        with open(os.path.join(conf, 'staging.yml'), 'w') as file:
            file.write(
                '''---
'''
            )
        with open(os.path.join(conf, 'prod.yml'), 'w') as file:
            file.write(
                '''---
'''
            )


@task
def docker(_, name, force=False):
    ''' create a config folder with dev, staging & prod yamls '''
    docker_file = 'Dockerfile'
    print(f'gen docker: {docker_file}')
    if os.path.isfile(docker_file) and force is False:
        print('that file already exists: ', docker_file)
    else:
        with open(docker_file, 'w') as file:
            file.write(
                '''# Use an official Python runtime as a parent image
FROM python:3.7-slim
RUN apt-get update && apt-get install -y git

# Set the working directory to /app
WORKDIR /app

# Copy the current directory contents into the container at /app
COPY . /app

# upgrade pip to make sure it can use git repos
RUN pip install --upgrade pip
RUN pip install bluemax

# Install any needed packages
RUN [ -f requirements.txt ] && pip install -r requirements.txt ||  echo "No requirements.txt"

# Make port 80 available to the world outside this container
EXPOSE 80

# Define environment variable
ENV PYTHONPATH ".:${PYTHONPATH}"
ENV PORT 80

'''
            )
        print(f'gen docker: docker-compose.yml')
        with open('docker-compose.yml', 'w') as file:
            file.write(
                f'''version: '3'
services:
  web:
    build: .
    image: {name}:latest
    container_name: {name}-web
    command: bluemax run -m {name}
    ports:
      - "8080:80"
'''
            )




""" We load our settings here """
from collections.abc import Mapping
import os
import importlib
import logging
import yaml

LOGGER = logging.getLogger(__name__)

SETTINGS = {}


def extend(extension, value):
    """ extension is a qname, resolved and run with value """
    mdl, func = extension.split(":")
    func = getattr(importlib.import_module(mdl), func)
    return func(value)


def merge_dict(dct, merge_dct):
    for key, val in merge_dct.items():
        if isinstance(dct.get(key), dict) and isinstance(val, Mapping):
            dct[key] = merge_dict(dct[key], val)
        else:
            dct[key] = val
    return dct


def init_settings(filename=None):
    """ load a config yaml file """
    filename = filename if filename else os.getenv("config", "settings.yml")
    if os.path.isfile(filename):
        LOGGER.info("loading settings from %s", filename)
        with open(filename, "r") as file:
            SETTINGS.update(yaml.safe_load(file))
    else:
        SETTINGS["settings"] = "local"  # stop rentrance


def get_settings(name, default=None):
    """
        access to settings - load is automatic
        if the settings cannot be found the getenv
        will be returned
    """
    if not SETTINGS:
        init_settings()
    return SETTINGS.get(name, os.getenv(name, default))
